/***************************************************************************
 *            mcp23017.c
 ****************************************************************************/

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor Boston, MA 02110-1301,  USA
 */


#include <avr/io.h>
#include <avr/pgmspace.h>
#include "config.h"

#if defined(USE_TWI)
#include <string.h> // memset

#include "usart.h"
#include "math.h"
#include "twi.h"
#include "mcp23017.h"
#include "timer.h"
#include "modbus_slave.h"

#define MCP23017_BASEADR 0x20
#define Mcp23017_MAX 8
#define Mcp23017_MAX_INPUT 4


//Register defines from data sheet - we set IOCON.BANK to 0
//as it is easier to manage the registers sequentially.
#define MCP23017_IODIR   0x00
#define MCP23017_IPOL    0x2
#define MCP23017_GPINTEN 0x4
#define MCP23017_DEFVAL  0x6
#define MCP23017_INTCON  0x8
#define MCP23017_IOCON   0x0A
#define MCP23017_GPPU    0x0C
#define MCP23017_INTF    0x0E
#define MCP23017_INTCAP  0x10
#define MCP23017_GPIO    0x12
#define MCP23017_OLAT    0x14


#define eMaxBits      16
#define eMaxSetCount  5 

typedef struct
{
    uint8_t  uiSetCounter[eMaxBits];
    uint16_t uiDataOut;
    uint16_t uiDataIn;
} TCheckSetBits;


unsigned short uiCheckBitSet(TCheckSetBits* pCheckSetBits)
{
    for( int i = 0; i < eMaxBits; i++ )
    {
        if( pCheckSetBits->uiDataIn & ( 1 << i ) )
        {
            pCheckSetBits->uiSetCounter[i] =
                ( pCheckSetBits->uiSetCounter[i] >= eMaxSetCount
                                                   ? eMaxSetCount
                                                   : pCheckSetBits->uiSetCounter[i] + 1 );
        }
        else
        {
            pCheckSetBits->uiSetCounter[i] = 0;
        }

        if( pCheckSetBits->uiSetCounter[i] >= eMaxSetCount )
        {
            pCheckSetBits->uiDataOut |= ( 1 << i ); // set bit
        }
        else
        {
            pCheckSetBits->uiDataOut &= ~( 1 << i ); // clear bit
        }
    }
    return pCheckSetBits->uiDataOut;
}




typedef struct 
{
    uint8_t  ucAdr;
    uint8_t  bOutput : 1;
    uint8_t  uiRegisterIndex : 4;
	uint16_t GPIO;
  
} tMcp23017Data;


tMcp23017Data g_Mcp23017Data        [ Mcp23017_MAX       ];
TCheckSetBits g_tInputRegister      [ Mcp23017_MAX_INPUT ];
uint8_t       g_uiInputRegisterUsed  = 0;
uint8_t       g_uiOutputRegisterUsed = 0;



char MCP23017_WriteByte(uint8_t I2CAddr, uint8_t RegAddr, uint8_t Value) 
{
    char bRet = TRUE;
    
    bRet = TWI_SendAddress( MCP23017_BASEADR + I2CAddr , TWI_WRITE );
    if( TRUE == bRet )
    {   
        bRet = TWI_Write( RegAddr );
        if( bRet )
        {
            bRet = TWI_Write( Value );
        }              
    }
    
    TWI_SendStop();
        
    return bRet;
} 


char MCP23017_WriteRegister(uint8_t I2CAddr, uint8_t RegAddr, uint16_t Value) 
{
    char bRet = TRUE;
    
    bRet = TWI_SendAddress( MCP23017_BASEADR + I2CAddr , TWI_WRITE );
    if( TRUE == bRet )
    {   
        bRet = TWI_Write( RegAddr );
        if( bRet )
        {
            bRet = TWI_Write( (Value >> 8) & 0xFF );
            if( bRet )
            {
                bRet = TWI_Write( Value & 0xFF );
            }
        }              
    }
    
    TWI_SendStop();
        
    return bRet;
} 


char MCP23017_ReadByte(uint8_t I2CAddr, uint8_t RegAddr, uint8_t *pRetValue) 
{
    char bRet = TRUE;
    char Byte = 0;
    
    bRet = TWI_SendAddress( MCP23017_BASEADR + I2CAddr , TWI_WRITE );
    if( TRUE == bRet )
    {   
            bRet = TWI_Write( RegAddr );              
    }
    TWI_SendStop();
    
    if( !bRet )
    {
        return FALSE;
    }
    
    
    bRet = TWI_SendAddress( MCP23017_BASEADR + I2CAddr , TWI_READ );
    if ( TRUE == bRet )
    {   
        bRet = TWI_ReadNack( &Byte );
        if( TRUE == bRet )
        {
            *pRetValue = Byte; 
        } 
    }
    else
    {
        bRet = FALSE;
    }
    
    TWI_SendStop();
    
    return bRet;
} 



char MCP23017_ReadRegister(uint8_t I2CAddr, uint8_t RegAddr, uint16_t *pwRetValue) 
{
    char bRet = TRUE;
    char Byte = 0;
    short wordRead = 0;
    
    
    bRet = TWI_SendAddress( MCP23017_BASEADR + I2CAddr , TWI_WRITE );
    if( TRUE == bRet )
    {   
            bRet = TWI_Write( RegAddr );              
    }
    TWI_SendStop();
    
    if( !bRet )
    {
        return FALSE;
    }
    
    
    bRet = TWI_SendAddress( MCP23017_BASEADR + I2CAddr , TWI_READ );
    if ( TRUE == bRet )
    {   
        bRet = TWI_ReadAck( &Byte );
        if( TRUE == bRet )
        {
            wordRead = ( ( Byte << 8 ) & 0xFF00 );  // HighByte
            
            bRet = TWI_ReadNack( &Byte );
            if( TRUE == bRet )
            {
                wordRead |= Byte; // LowByte
            }
        } 
    }
    else
    {
        bRet = FALSE;
    }
    
    TWI_SendStop();
    
    if( TRUE == bRet )
    {
        *pwRetValue = wordRead;
    }
    
    return bRet;
} 



char MCP23017_bInit( void )
{
    uint8_t  i    = 0;
    uint8_t  bRet = FALSE;
	
    usart_write( "Init MCP23017 port extension\n\r" );
    for ( i = 0 ; i < Mcp23017_MAX_INPUT ; i++ )
	{
        memset( &g_tInputRegister[i], 0, sizeof(TCheckSetBits) );
    }
	for ( i = 0 ; i < Mcp23017_MAX ; i++ )
	{
        memset( &g_Mcp23017Data[i], 0, sizeof(tMcp23017Data) );
        //g_Mcp23017Data[i].ucAdr = 0;
        //g_Mcp23017Data[i].bOutput = FALSE;
        
        
        //Set the IOCON.BANK bit to 0 to enable sequential addressing
        //IOCON 'default' address is 0x05, but will
        //change to our definition of IOCON once this write completes.
        if( MCP23017_WriteByte( MCP23017_BASEADR + i, 0x5, 0x0) ) 
		{
            bRet = TRUE;
            g_Mcp23017Data[i].ucAdr = i;
            
            if( 0 == i%2 )
            {
                //even address - input module
                usart_write( "MCP23017 Input: 0x%02x\r\n", MCP23017_BASEADR + i );
                g_uiInputRegisterUsed++;
                if( g_uiInputRegisterUsed > Mcp23017_MAX_INPUT )
                {
                    g_uiInputRegisterUsed --;
                    usart_write( "ERROR MCP23017: Too many input registers\r\n" );
                }
                else
                {
                    g_Mcp23017Data[i].uiRegisterIndex = g_uiInputRegisterUsed;
                }
            }
            else
            {
                g_Mcp23017Data[i].bOutput = TRUE;
                g_uiOutputRegisterUsed++;
                g_Mcp23017Data[i].uiRegisterIndex = g_uiOutputRegisterUsed;
                
                //odd address - output module
                usart_write( "MCP23017 Output: 0x%02x\r\n", MCP23017_BASEADR + i );
                
                // Port A and B all bits output
                MCP23017_WriteRegister( MCP23017_BASEADR + i, MCP23017_IODIR, 0x0);
            }
		}
	}
    Modbus_SetRegister( MODBUS_REG_OUT_MCP23017_INPUTS , g_uiInputRegisterUsed  );
    Modbus_SetRegister( MODBUS_REG_OUT_MCP23017_OUTPUTS, g_uiOutputRegisterUsed );
    
    return bRet;
}

void MCP23017_Update( void )
{
    for ( int i = 0 ; i < Mcp23017_MAX ; i++ )
	{
        if( 0 == g_Mcp23017Data[i].ucAdr )
        {
            break;
        }
        
        
        if( TRUE == g_Mcp23017Data[i].bOutput )
        {
            // OUTPUT
            g_Mcp23017Data[i].GPIO = Modbus_GetRegister( 
                MODBUS_REGISTER_MCP32017_START + g_Mcp23017Data[i].uiRegisterIndex - 1 );
            MCP23017_WriteRegister( MCP23017_BASEADR + g_Mcp23017Data[i].ucAdr, 
                                    MCP23017_GPIO, 
                                    g_Mcp23017Data[i].GPIO); 
        }
        else
        {
            // INPUT
            MCP23017_ReadRegister(  MCP23017_BASEADR + g_Mcp23017Data[i].ucAdr, 
                                    MCP23017_GPIO, 
                                    &g_Mcp23017Data[i].GPIO); 
            if( 0 != g_Mcp23017Data[i].uiRegisterIndex )
            {
                // filter inputs
                g_tInputRegister[g_Mcp23017Data[i].uiRegisterIndex].uiDataIn =
                    g_Mcp23017Data[i].GPIO;
                g_Mcp23017Data[i].GPIO =
                    uiCheckBitSet( &g_tInputRegister[g_Mcp23017Data[i].uiRegisterIndex] );
                Modbus_SetRegister( 
                    MODBUS_REGISTER_MCP32017_START + g_Mcp23017Data[i].uiRegisterIndex - 1, 
                    g_Mcp23017Data[i].GPIO );
            }
        }
	}
}


/*
    int i = 0;
	int Temp = 0;
	char VZ = '+';
 
	for ( i = 0 ; i < 128 ; i++ )
	{
		if ( TWI_SendAddress( i , TWI_WRITE ) == TRUE )
		{
			usart_write( " 0x%02x\r\n", i );
		}			
		TWI_SendStop();
	}

	usart_write( "\r\nScanne nach LM75 Temperatursensoren:\r\n"); 

	for ( i = 0x48 ; i < 0x50 ; i++ )
	{
		if ( TWI_SendAddress( i , TWI_WRITE ) == TRUE )
		{
			TWI_Write( 0 );
			TWI_SendStop();

			TWI_SendAddress( i, TWI_READ );
			Temp = ( TWI_ReadAck() << 8 );
			Temp |= ( TWI_ReadNack() );
			TWI_SendStop();

			if ( Temp < 0 )
				VZ = '-';
			else
				VZ = '+';
			
			usart_write( "LM75 Sensor gefunden auf Adresse: 0x%02x, Temperatur %c%d.%01d ", i, VZ, ABS(Temp / 256) , ABS((Temp << 8 ) / 6250 )); // 0.1 Grad C			
//			printf_P( PSTR("LM75 Sensor gefunden auf Adresse: 0x%02x, Temperatur %d.%c C\r\n"), i, Temp >> 8 , NK );
		}			
	}
   */



#endif
