/*
File: modbus_slave.c
*/

#include <string.h>
#include "modbus_slave.h"

#if USE_MODBUS

volatile uint16_t Modbus_InputRegister [MODBUS_REGISTERS];
volatile uint16_t Modbus_OutputRegister[MODBUS_REGISTERS];


void Modbus_InitProtocol(void)
{
    MODBUS_DEBUG("Init Modbus\r\n");
     
    // initialize internal modbus register
    memset(&Modbus_InputRegister , 0, sizeof(Modbus_InputRegister) );
    memset(&Modbus_OutputRegister, 0, sizeof(Modbus_OutputRegister) );
     
     //Port in Anwendungstabelle eintragen fr eingehende Modbus Daten!
    add_tcp_app (MODBUS_TCP_IP_PORT, (void(*)(unsigned char))Modbus_ProcessRequest);
     
    return;
}


int Modbus_CheckCode(uint8_t *buffer)
{
    if(    buffer[MODBUS_CODE_INDEX] == CODE_READ_MULT_REGISTERS 
        || buffer[MODBUS_CODE_INDEX] == CODE_WRITE_MULT_REGISTERS ) 
    {
        return MODBUS_OK;
    }

    return MODBUS_ERROR;
}

int Modbus_CheckQuan(uint8_t *buffer)
{
    int quantity_regs = (buffer[MODBUS_QUAN_HI_INDEX]<<8) + buffer[MODBUS_QUAN_LO_INDEX];
    
    switch (buffer[MODBUS_CODE_INDEX])
    {
        case CODE_READ_MULT_REGISTERS:
            if ((quantity_regs < 1) || (quantity_regs > 0x7d)) 
                { return MODBUS_ERROR; }
            break; 

        case CODE_WRITE_MULT_REGISTERS:
            if ((quantity_regs < 1) || (quantity_regs > 0x7b)) 
                { return MODBUS_ERROR; }
            if (buffer[MODBUS_BYTE_CNT_INDEX16] != (quantity_regs*2)) 
                { return MODBUS_ERROR; }
            break;
    }
    return MODBUS_OK;
}

int Modbus_CheckAddress(uint8_t *buffer)
{
    int start_addr    = (buffer[MODBUS_ADDR_HI_INDEX]<<8) + buffer[MODBUS_ADDR_LO_INDEX];
    int quantity_regs = (buffer[MODBUS_QUAN_HI_INDEX]<<8) + buffer[MODBUS_QUAN_LO_INDEX];
    if ((start_addr + quantity_regs) > MODBUS_REGISTERS) 
    {
        return MODBUS_ERROR;
    }
    return MODBUS_OK;
}

int Modbus_BuildResponse(uint8_t *buffer)
{
    int16_t i,j;
    int16_t n_bytes = 0;
      
    int start_addr    = (buffer[MODBUS_ADDR_HI_INDEX]<<8) + buffer[MODBUS_ADDR_LO_INDEX];
    int quantity_regs = (buffer[MODBUS_QUAN_HI_INDEX]<<8) + buffer[MODBUS_QUAN_LO_INDEX];  
    switch(buffer[MODBUS_CODE_INDEX])
    {
        // address checking has been already done in Modbus_CheckAddress()
        case CODE_READ_MULT_REGISTERS:
            buffer[MODBUS_BYTE_CNT_INDEX ] = 3+2*quantity_regs;
            buffer[MODBUS_BYTE_CNT2_INDEX] = 2*quantity_regs;
            for (j=0,i=start_addr;j<quantity_regs;i++,j++)
            {   
                buffer[MODBUS_DATA_BASE_INDEX3+(2*j)  ] = Modbus_OutputRegister[i]>>8;
                buffer[MODBUS_DATA_BASE_INDEX3+(2*j)+1] = Modbus_OutputRegister[i]&0x00ff;
            }
            n_bytes = (9+2*quantity_regs);
            break;

        case CODE_WRITE_MULT_REGISTERS:
            buffer[MODBUS_BYTE_CNT_INDEX]=6;
            for (j=0,i=start_addr;j<quantity_regs;i++,j++)
            {
                Modbus_InputRegister[i]=(buffer[MODBUS_DATA_BASE_INDEX16+(2*j)]<<8)+buffer[MODBUS_DATA_BASE_INDEX16+(2*j)+1];
            }
            n_bytes = 12;  
            break;
    }
    return n_bytes;
}

void Modbus_SendException(uint8_t index, uint8_t *buffer, uint8_t exc_code)
{
    MODBUS_DEBUG("Modbus Exception\r\n");

    buffer[MODBUS_BYTE_CNT_INDEX]   = 3;
    buffer[MODBUS_CODE_INDEX]       = buffer[MODBUS_CODE_INDEX] | 0x80;
    buffer[MODBUS_EXCEP_CODE_INDEX] = exc_code;
    
    tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
    create_new_tcp_packet(9,index);
	tcp_entry[index].time = TCP_MAX_ENTRY_TIME;
}

void Modbus_ProcessRequest(uint8_t index)
{
    MODBUS_DEBUG("Modbus Request\r\n");

    //connection closed
    if (tcp_entry[index].status & FIN_FLAG)
    {
        MODBUS_DEBUG("Modbus close\r\n");
        return;
    }

    if (tcp_entry[index].app_status == 0 || tcp_entry[index].app_status == 1)
	{
        // connection opened
		tcp_entry[index].app_status = 2; 
        MODBUS_DEBUG("Modbus open\r\n");
	}
    
	if ((tcp_entry[index].app_status >= 1) && (tcp_entry[index].status&PSH_FLAG))
	{
		tcp_entry[index].app_status = 2;	
		
        uint8_t* buffer = &eth_buffer[TCP_DATA_START_VAR];

        // unsigned char chSlave = buffer[MODBUS_ID_INDEX];
 
        if (Modbus_CheckCode(buffer) == MODBUS_ERROR)
        {
           return (Modbus_SendException(index, buffer, MODBUS_CODE_EXCEPTION));
        }
        if (Modbus_CheckQuan(buffer) == MODBUS_ERROR)
        {
           return (Modbus_SendException(index, buffer, MODBUS_QUAN_EXCEPTION));
        }


        if (Modbus_CheckAddress(buffer) == MODBUS_ERROR)
        {
           return (Modbus_SendException(index, buffer, MODBUS_ADDR_EXCEPTION));
        }

        // build answer
        unsigned int count = Modbus_BuildResponse(buffer);

        // send answer
        tcp_entry[index].status =  ACK_FLAG | PSH_FLAG;
		create_new_tcp_packet(count,index);
		tcp_entry[index].time = TCP_MAX_ENTRY_TIME;
	}
		
	return; 
}

inline void Modbus_SetRegister(uint8_t index, uint16_t value)
{
    if(index >= MODBUS_REGISTERS)
    {
        return;
    }
    Modbus_OutputRegister[index] = value;
}

inline uint16_t Modbus_GetRegister(uint8_t index)
{
    if(index >= MODBUS_REGISTERS)
    {
        return 0;
    }
    return Modbus_InputRegister[index];
}



#endif //USE_MODBUS


